# This file was auto-generated by salt's setup on Wednesday, 24 October 2018 @ 14:10:31 UTC.

from salt.version import SaltStackVersion

__saltstack_version__ = SaltStackVersion(2016, 11, 10, 0, '', 0, 0, None)
